#!/usr/bin/perl

#############################################################################
# Copyright (C) 2023, Matrix Science Limited.                               #
#                                                                           #
# This script is free software. You can redistribute and/or                 #
# modify it under the terms of the GNU General Public License               #
# as published by the Free Software Foundation; either version 2            #
# of the License or, (at your option), any later version.                   #
#                                                                           #
# These modules are distributed in the hope that they will be useful,       #
# but WITHOUT ANY WARRANTY; without even the implied warranty of            #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              #
# GNU General Public License for more details.                              #
#############################################################################

use v5.14;
use warnings;

use XML::Twig;

sub usage {
say STDERR <<USAGE;
Usage: $0 F001234.xml

This script summarises the chimericity of the peak lists and search results,
given a file of Mascot search results exported in Mascot XML format.

Creating a suitable XML file:

- Open the search results in Mascot Server.
- Click on Export and choose XML format.
- Set desired target FDR and enable/disable Percolator.
- Accept the defaults under Search Information.
- Accept the defaults under Protein Hit Information.
- Accept the defaults under Peptide Match Information.
- Enable Query Level Information and tick "Query title, retention
  time*, scan number*"

WARNING: Mascot Distiller multi-file projects are supported, but the
detection depends on the query title string. If the peak lists come
from a different peak picking tool that merges multiple raw files into
a single MGF file, the counts will not be correct.

USAGE
}

my $mascot_xml = shift @ARGV;

if (not defined $mascot_xml or not -e $mascot_xml) {
    exit usage();
}

process($mascot_xml, \my %counts);
summarise(\%counts);

exit 0;

BEGIN {
    my %warnings_seen;

    sub _warn_once {
        my ($str) = @_;
        if (not $warnings_seen{$str}) {
            say STDERR 'WARNING: ', $str;
            $warnings_seen{$str} = 1;
        }
    }
}

sub process {
    my ($filepath, $counts) = @_;

    # Call this progress reporter in each handler to periodically
    # print percentage progress.
    my $file_size = -s $filepath;
    my $progress_reporter = sub {};

    if ($file_size) {
        my @percentage_points = map { $_/20 } 0..19;

        $progress_reporter = sub {
            my ($twig) = @_;
            my $pos = $twig->current_byte;
            if (@percentage_points) {
                my $here = $pos / $file_size;
                if ($here >= $percentage_points[0]) {
                    shift @percentage_points;
                    say STDERR sprintf("PROGRESS: %.0f%%", $here * 100);
                }
            }
        };
    }

    my $sigthreshold = 0;
    my %pep_counts;
    my @query_map;

    my $process_msr_sigthreshold = sub {
        # The <sigthreshold> element always precedes peptide and query
        # data, so we can assume this assignment always happens before
        # anything else is called.
        my ($t, $elt) = @_;

        $progress_reporter->($t);

        $sigthreshold = $elt->text;

        $elt->purge;
    };

    my $process_msr_peptide = sub {
        my ($t, $msr_peptide) = @_;

        $progress_reporter->($t);

        my $query = $msr_peptide->att('msr:query');
        my $rank = $msr_peptide->att('msr:rank');

        if (not defined $rank or $rank eq '') {
            # Unmatched spectra encoded as <u_peptide> elements have no rank.
            return;
        }

        # Only count rank 1 matches.
        return if $rank > 1;

        # The protein hit can contain both significant and
        # non-significant matches (it's an option in the
        # export form). Check the expect value against the
        # significance threshold.
        #
        # If this is an unassigned match, we know expect value is
        # already non-significant.
        my $expect = $msr_peptide->first_child('msr:pep_expect');

        if ($expect) {
            $expect = $expect->text;

            # Note: ET matches only have an expect value in
            # Mascot Server 2.8 and later. The second pass
            # significance threshold can be different from the
            # first pass threshold, and it's currently not
            # available in the XML. The best that can be done is
            # emit a warning if this is an error tolerant search
            # and the expect value is defined.
            if (not defined $expect or $expect eq '' or $expect <= 0.0) {
                _warn_once('Missing expect value in <pep_expect> element -- error tolerant match? Treating it as "not significant" for counting purposes.');
                $expect = 1e99;
            }

        } else {
            _warn_once('Missing element <pep_expect> -- please tick "Expectation value" under Peptide Match Information when exporting the XML.');
            $expect = 1e99; # not significant
        }

        # We may encounter the same query multiple times, e.g.
        # sequence is shared between proteins, but each time
        # the rank 1 match is either significant or it isn't.
        if ($expect <= $sigthreshold) {
            ++$pep_counts{$query}{'matched'};
        } else {
            ++$pep_counts{$query}{'unassigned'};
        }
        
        $msr_peptide->purge;
    };

    my $process_msr_query = sub {
        my ($t, $msr_query) = @_;

        $progress_reporter->($t);

        my $number = $msr_query->att('msr:number');

        # If the MGF is from a Distiller multi-file project,
        # the SCANS field in the Mascot XML is not a reliable
        # indicator. For example, SCANS[1]=3 and SCANS[0]=3
        # are both reported as <SCANS>3</SCANS>.
        #
        # Note also that SCANS could be a list of scan numbers
        # if any MS/MS spectra have been summed. Use it as an
        # opaque ID.
        my $SCANS = $msr_query->first_child('msr:SCANS');
        if ($SCANS) {
            $SCANS = $SCANS->text;
        } else {
            _warn_once('Missing <SCANS> element in <queries> section -- cannot map Mascot query numbers to scan numbers.');
            $SCANS = $number; # fudge it
        }

        # There is currently no reliable way to detect
        # Distiller multi-file project, other than to look at
        # the query title field.
        my $title = $msr_query->first_child('msr:StringTitle');
        if (defined $title) {
            if ($title->text =~ m{Scan +\d+.*? +from +file +\[(\d+)\]}i) {
                my $file_id = $1;
                _warn_once("Scan title contains \"from file [$file_id]\" -- assuming this is a Distiller multi-file project.");
                $SCANS = join("\0", $file_id, $SCANS);
            }
        } else {
            #_warn_once('Missing <StringTitle> element in <queries> section.');
        }

        # The precursor m/z of a Mascot query is one of the
        # PEPMASS values in the chimeric input spectrum.
        my $precursor_mz = $msr_query->first_child('msr:query_moverz');
        if ($precursor_mz) {
            $precursor_mz = $precursor_mz->text;
        } else {
            _warn_once("Missing <query_moverz> element in <query> $number -- setting precursor m/z to 0.");
            $precursor_mz = 0;
        }

        $query_map[$number] = {
            scan_id => $SCANS,
            precursor_mz => $precursor_mz,
        };

        $msr_query->purge;
    };

    my $twig = XML::Twig->new(
        map_xmlns => {
            'http://www.matrixscience.com/xmlns/schema/mascot_search_results_2' => 'msr',
        },

        twig_handlers => {
            'msr:format_parameters/msr:sigthreshold' => $process_msr_sigthreshold,

            'msr:hits/msr:hit/msr:protein/msr:peptide' => $process_msr_peptide,
            'msr:unassigned/msr:u_peptide' => $process_msr_peptide,

            'msr:queries/msr:query' => $process_msr_query,
        },
    );

    $twig->parsefile($filepath);

    if (not @query_map) {
        _warn_once('XML file is missing <queries> section -- cannot count anything.');
        return 0;
    }

    my %scan_matches;

    for my $q (keys %pep_counts) {
        my $scan_id = $query_map[$q]{'scan_id'};
        my $precursor_mz = $query_map[$q]{'precursor_mz'};

        # Count significant queries and scans.
        #
        # For scans, we want a breakdown: if the scan has n=3 PEPMASS
        # lines, we want to know how many of those (0..n) has a
        # significant match.
        if ($pep_counts{$q}{'matched'}) {
            $$counts{'num_significant_queries'} += 1;

            ++$scan_matches{$scan_id};
        }
    }

    my %scan_PEPMASS;

    for my $q (1 .. $#query_map) {
        ++$counts{'num_queries'};

        my $scan_id = $query_map[$q]{'scan_id'};
        my $precursor_mz = $query_map[$q]{'precursor_mz'};

        # Reconstruct the number of PEPMASS lines in the input MGF for
        # each scan.
        ++$scan_PEPMASS{$scan_id}{$precursor_mz};

        # Count unassigned/unmatched queries.
        if (not $pep_counts{$q}{'matched'}) {
            $$counts{'num_unassigned_queries'} += 1;
        }
    }

    for my $scan_id (keys %scan_PEPMASS) {
        ++$$counts{'num_scans'};

        my $num_PEPMASS = scalar(keys %{ $scan_PEPMASS{$scan_id} });
        ++$$counts{'PEPMASS_count'}{$num_PEPMASS};

        my $num_matched = $scan_matches{$scan_id} // 0;

        if ($num_matched > 0) {
            $$counts{'num_significant_scans'} += 1;
        } else {
            $$counts{'num_unassigned_scans'} += 1;
        }

        ++$counts{'scan_breakdown'}{$num_PEPMASS}{$num_matched};
    }

    return 1;
}

sub summarise {
    my ($counts) = @_;

    my $comma = ', ';

    say '---Peak list chimericity report:';
    say '';
    say join($comma, 'Number of Mascot queries', $$counts{'num_queries'} // 0);
    say join($comma, 'Number of MS/MS scans', $$counts{'num_scans'} // 0);

    say '';
    say join($comma, 'Precursor masses per scan', 'Count');
    for my $p (sort { $a <=> $b } keys %{ $$counts{'PEPMASS_count'} }) {
        say join($comma, $p, $$counts{'PEPMASS_count'}{$p} // 0);
    }

    say '';
    say '---Database search chimericity report:';
    say '';

    say join($comma, 'Number of significant PSMs', $$counts{'num_significant_queries'} // 0);
    say join($comma, 'Number of unassigned/unmatched queries', $$counts{'num_unassigned_queries'} // 0);
    say '';
    say join($comma, 'Number of scans with at least one significant match', $$counts{'num_significant_scans'} // 0);
    say join($comma, 'Number of unassigned/unmatched scans', $$counts{'num_unassigned_scans'} // 0);

    for my $p (sort {$a <=> $b } keys %{ $$counts{'scan_breakdown'} }) {
        say '';

        if ($p == 1) {
            say 'Scans with 1 precursor mass (non-chimeric):';
        } else {
            say sprintf("Scans with %d precursor masses", $p);
        }

        say '';
        say join($comma, 'Significant matches per scan', 'Count');
        for my $i (0 .. $p) {
            say join($comma, $i, $$counts{'scan_breakdown'}{$p}{$i} // 0);
        }
    }
    
}
